package dev.struchkov.godfather.quarkus.core;

import dev.struchkov.godfather.main.domain.StorylineHistory;
import dev.struchkov.godfather.main.domain.content.Message;
import dev.struchkov.godfather.main.domain.unit.UnitActiveType;
import dev.struchkov.godfather.quarkus.context.service.ErrorHandler;
import dev.struchkov.godfather.quarkus.context.service.Modifiable;
import dev.struchkov.godfather.quarkus.core.action.ActionUnit;
import dev.struchkov.godfather.quarkus.core.service.StorylineService;
import dev.struchkov.godfather.quarkus.domain.unit.MainUnit;
import dev.struchkov.godfather.quarkus.domain.unit.UnitRequest;
import dev.struchkov.haiti.context.exception.NotFoundException;
import io.smallrye.mutiny.Multi;
import io.smallrye.mutiny.Uni;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;

import static dev.struchkov.haiti.utils.Checker.checkEmpty;
import static dev.struchkov.haiti.utils.Checker.checkNotNull;
import static java.util.Collections.emptyList;

public class GeneralAutoResponder<M extends Message> {

    protected final StorylineService<M> storyLineService;
    protected Map<String, List<ActionUnit>> actionUnitMap = new HashMap<>();
    protected List<Modifiable<M>> modifiable;
    protected ErrorHandler errorHandler;

    protected GeneralAutoResponder(StorylineService<M> storyLineService) {
        this.storyLineService = storyLineService;
    }

    public void registrationActionUnit(ActionUnit actionUnit) {
        actionUnitMap.computeIfAbsent(actionUnit.getUnitType(), k -> new ArrayList<>());
        actionUnitMap.get(actionUnit.getUnitType()).add(actionUnit);
    }

    public void initModifiable(List<Modifiable<M>> modifiable) {
        this.modifiable = modifiable;
    }

    /**
     * Позволяет установить перехватчик и обработчик исключений, возникающих при обработке юнитов.
     */
    public void setErrorHandler(ErrorHandler errorHandler) {
        this.errorHandler = errorHandler;
    }

    public void setDefaultUnit(String unitName) {
        storyLineService.setDefaultUnit(unitName);
    }

    public Uni<Void> processingNewMessage(M newMessage) {
        return Uni.createFrom().item(newMessage)
                .onItem().ifNotNull().transformToUni(
                        message -> processing(newMessage)
                )
                .replaceWithVoid();
    }

    public Uni<Void> processingNewMessages(List<M> newMessages) {
        return Uni.createFrom().item(newMessages)
                .onItem().ifNotNull().transformToUni(
                        messages -> {
                            if (checkEmpty(newMessages)) return Uni.createFrom().voidItem();
                            return Multi.createFrom().iterable(newMessages)
                                    .onItem().transform(this::processing)
                                    .collect().asList().replaceWithVoid();
                        }
                );
    }

    private Uni<Void> processing(M message) {
        return Uni.createFrom().item(message)
                .onItem().ifNotNull().transform(m -> modifiable)
                .replaceIfNullWith(emptyList())
                .onItem().transformToMulti(modifiables -> Multi.createFrom().iterable(modifiables))
                .onItem().transformToUni(mModifiable -> mModifiable.change(message))
                .concatenate().toUni().replaceWith(
                        storyLineService.getNextUnitByPersonId(message.getFromPersonId())
                                .onItem().ifNotNull().transformToUni(nextUnits -> Responder.nextUnit(message, nextUnits))
                                .onItem().ifNull().switchTo(Uni.createFrom().optional(storyLineService.getDefaultUnit()))
                                .onItem().ifNotNull().transformToUni(answerUnit -> answer(UnitRequest.of(answerUnit, message)))
                );
    }

    public Uni<Void> answer(UnitRequest<MainUnit, M> unitRequest) {
        return getAction(unitRequest)
                .chain(this::activeUnitAfter)
                .onFailure().call(
                        throwable -> {
                            if (checkNotNull(errorHandler)) {
                                return errorHandler.handle(unitRequest.getMessage(), throwable);
                            }
                            return Uni.createFrom().voidItem();
                        }
                )
                .onFailure(th -> checkNotNull(errorHandler)).recoverWithNull()
                .replaceWithVoid();
    }

    private Uni<UnitRequest<MainUnit, M>> activeUnitAfter(UnitRequest<MainUnit, M> unitRequest) {
        final Set<MainUnit<M>> nextUnits = unitRequest.getUnit().getNextUnits();
        if (checkNotNull(nextUnits)) {
            Optional<MainUnit<M>> first = nextUnits.stream()
                    .filter(unit -> UnitActiveType.AFTER.equals(unit.getActiveType()))
                    .findFirst();
            if (first.isPresent()) {
                return Uni.createFrom().voidItem().onItem().transformToUni(
                                v -> getAction(UnitRequest.of(first.get(), unitRequest.getMessage()))
                        )
                        .onItem().transformToUni(
                                (Function<UnitRequest<MainUnit, M>, Uni<? extends UnitRequest<MainUnit, M>>>) newRequest -> {
                                    return activeUnitAfter(UnitRequest.of(newRequest.getUnit(), newRequest.getMessage()));
                                }
                        );
            }
        }
        return Uni.createFrom().item(unitRequest);
    }

    private Uni<UnitRequest<MainUnit, M>> getAction(UnitRequest<MainUnit, M> unitRequest) {
        final MainUnit<M> unit = unitRequest.getUnit();
        final M message = unitRequest.getMessage();
        final String typeUnit = unit.getType();
        if (actionUnitMap.containsKey(typeUnit)) {
            List<ActionUnit> actionUnits = actionUnitMap.get(typeUnit);
            final Class<? extends Message> messageClass = message.getClass();
            final ActionUnit<MainUnit, M> actionUnit = getActionUnit(actionUnits, messageClass);
            if (checkNotNull(actionUnit)) {
                return Uni.createFrom().deferred(() -> actionUnit.action(unitRequest))
                        .flatMap(
                                newUnitRequest -> {
                                    final Optional<MainUnit<M>> optDefaultUnit = storyLineService.getDefaultUnit();
                                    if (!unit.isNotSaveHistory() && (optDefaultUnit.isEmpty() || !optDefaultUnit.get().equals(unit))) {
                                        return Uni.combine().all().unis(
                                                Uni.createFrom().item(newUnitRequest),
                                                storyLineService.save(StorylineHistory.builder()
                                                        .personId(message.getFromPersonId())
                                                        .unitName(unit.getName())
                                                        .message(message)
                                                        .build())
                                        ).asTuple();
                                    }
                                    return Uni.combine().all().unis(Uni.createFrom().item(newUnitRequest), Uni.createFrom().voidItem()).asTuple();
                                }
                        ).onItem().transformToUni(
                                t -> {
                                    final UnitRequest<MainUnit, M> newUnitRequest = t.getItem1();
                                    final MainUnit<M> newUnit = newUnitRequest.getUnit();
                                    return !unit.equals(newUnit) ? getAction(newUnitRequest) : Uni.createFrom().item(unitRequest);
                                }
                        );
            }
        }
        throw new NotFoundException("ActionUnit для типа {0} не зарегистрирован", unit.getType());
    }

    private ActionUnit<MainUnit, M> getActionUnit(List<ActionUnit> actionUnits, Class<? extends Message> messageClass) {
        for (ActionUnit actionUnit : actionUnits) {
            if (actionUnit.getMessageType().isAssignableFrom(messageClass)) {
                return actionUnit;
            }
        }
        return null;
    }

}
