package dev.struchkov.godfather.quarkus.core;

import dev.struchkov.godfather.main.domain.content.Message;
import dev.struchkov.godfather.main.domain.unit.KeyWord;
import dev.struchkov.godfather.quarkus.core.util.UnitPriorityComparator;
import dev.struchkov.godfather.quarkus.domain.unit.MainUnit;
import dev.struchkov.godfather.quarkus.domain.unit.func.UniPredicate;
import io.smallrye.mutiny.Multi;
import io.smallrye.mutiny.Uni;
import lombok.extern.slf4j.Slf4j;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import static dev.struchkov.godfather.main.core.utils.Parser.splitWords;
import static dev.struchkov.haiti.utils.Checker.checkEmpty;
import static dev.struchkov.haiti.utils.Checker.checkNotEmpty;
import static dev.struchkov.haiti.utils.Checker.checkNotNull;
import static dev.struchkov.haiti.utils.Checker.checkNull;
import static dev.struchkov.haiti.utils.Exceptions.utilityClass;
import static dev.struchkov.haiti.utils.Inspector.isNotNull;
import static java.lang.Boolean.TRUE;

/**
 * Реализуют основную логику автоответчика.
 *
 * @author upagge [07/07/2019]
 */
@Slf4j
public final class Responder {

    /**
     * Компаратор для сортировки Unit-ов
     */
    private static final UnitPriorityComparator UNIT_PRIORITY_COMPARATOR = new UnitPriorityComparator();

    private Responder() {
        utilityClass();
    }

    /**
     * Выбирает, какой Unit будет отдан для обработки
     *
     * @param nextUnits Множество следующих Unit-ов
     * @param message   Запрос пользователя - текстовое сообщение
     * @return Юнит, который нуждается в обработке в соответствии с запросом пользователя
     */
    public static <M extends Message> Uni<MainUnit<M>> nextUnit(M message, Collection<MainUnit<M>> nextUnits) {
        isNotNull(nextUnits);
        if (checkNotEmpty(nextUnits)) {
            return Multi.createFrom().iterable(nextUnits)
                    .onItem().transformToUniAndMerge(unit -> {
                        final String text = message.getText();
                        if (checkNotNull(text)) {
                            final Set<String> unitPhrases = unit.getPhrases();
                            final Set<Pattern> patterns = unit.getTriggerPatterns();

                            final boolean triggerPhrases = checkNotEmpty(unitPhrases) && unitPhrases.contains(text);
                            final boolean triggerPatterns = checkNotEmpty(patterns) && patterns.stream().anyMatch(pattern -> patternReg(pattern, text));
                            final boolean triggerKeyWords = percentageMatch(unit, text) >= unit.getMatchThreshold();
                            final boolean trigger = triggerPhrases || triggerPatterns || triggerKeyWords;

                            if (trigger) {
                                return Uni.createFrom().item(unit);
                            }
                        }

                        final UniPredicate<M> triggerCheck = unit.getTriggerCheck();
                        if (checkNotNull(triggerCheck)) {
                            return triggerCheck.test(message)
                                    .onItem().transform(matchesTrigger -> TRUE.equals(matchesTrigger) ? unit : null);
                        }

                        return Uni.createFrom().nullItem();
                    })
                    .collect().asList()
                    .onItem().transform(searchUnit -> {
                        if (searchUnit.isEmpty()) {
                            return nextUnits.stream().filter(Responder::isNotTrigger).collect(Collectors.toList());
                        } else {
                            return searchUnit;
                        }
                    })
                    .onItem().transform(list -> list.stream().max(UNIT_PRIORITY_COMPARATOR).orElse(null));
        } else {
            return Uni.createFrom().nullItem();
        }
    }

    private static <M extends Message> boolean isNotTrigger(MainUnit<M> nextUnit) {
        return isNotPattern(nextUnit) && isNotKeyWords(nextUnit) && isNotPhrase(nextUnit) && isNotCheck(nextUnit);
    }

    private static <M extends Message> boolean isNotCheck(MainUnit<M> unit) {
        return checkNull(unit.getTriggerCheck());
    }

    private static <M extends Message> boolean isNotPhrase(MainUnit<M> unit) {
        return checkEmpty(unit.getPhrases());
    }

    private static <M extends Message> boolean isNotPattern(MainUnit<M> unit) {
        return checkEmpty(unit.getTriggerPatterns());
    }

    private static <M extends Message> boolean isNotKeyWords(MainUnit<M> unit) {
        return checkEmpty(unit.getTriggerWords());
    }

    private static boolean patternReg(Pattern pattern, String message) {
        isNotNull(pattern);
        return message.matches(pattern.pattern());
    }

    private static <M extends Message> double percentageMatch(MainUnit<M> unit, String message) {
        final Set<KeyWord> unitKeyWords = unit.getTriggerWords();
        if (checkNotEmpty(unitKeyWords)) {
            final Set<String> messageWords = splitWords(message);
            final Set<KeyWord> intersection = getIntersection(unitKeyWords, messageWords);
            final double intersectionWeight = getIntersectionWeight(intersection);
            log.trace("Ключевые слова юнита: {} ({})", unitKeyWords, unitKeyWords.size());
            log.trace("Слова, которые прислал пользователь: {}", messageWords);
            log.trace("Пересечение: {} ({})", intersection, intersectionWeight);
            log.trace("Процент: {} Необходимо: {}", intersectionWeight / unitKeyWords.size() * 100.0, unit.getMatchThreshold());
            return (double) intersection.size() / (double) unitKeyWords.size() * 100.0;
        } else {
            return 0.0;
        }
    }

    private static double getIntersectionWeight(Set<KeyWord> intersection) {
        return intersection.stream().mapToInt(KeyWord::getImportant).sum();
    }

    private static Set<KeyWord> getIntersection(Set<KeyWord> unitKeyWords, Set<String> messageWords) {
        final Set<KeyWord> intersection = new HashSet<>();
        for (KeyWord unitKeyWord : unitKeyWords) {
            if (messageWords.contains(unitKeyWord.getWord())) {
                intersection.add(unitKeyWord);
            }
        }
        return intersection;
    }

}
