package dev.struchkov.godfather.quarkus.core.service;

import dev.struchkov.godfather.exception.ConfigAppException;
import dev.struchkov.godfather.main.domain.StorylineHistory;
import dev.struchkov.godfather.main.domain.UnitPointer;
import dev.struchkov.godfather.main.domain.content.Mail;
import dev.struchkov.godfather.quarkus.context.repository.StorylineHistoryRepository;
import dev.struchkov.godfather.quarkus.context.service.UnitPointerService;
import dev.struchkov.godfather.quarkus.core.Storyline;
import dev.struchkov.godfather.quarkus.core.StorylineFactory;
import dev.struchkov.godfather.quarkus.domain.unit.MainUnit;
import io.smallrye.mutiny.Uni;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.Optional;
import java.util.Set;

import static dev.struchkov.haiti.utils.Checker.checkEmpty;
import static dev.struchkov.haiti.utils.Checker.checkNotEmpty;
import static dev.struchkov.haiti.utils.Checker.checkNotNull;
import static dev.struchkov.haiti.utils.Inspector.isNotNull;

/**
 * Отвечает за работу со сценарием в личных сообщениях с пользователем.
 */
public class StorylineMailService implements StorylineService<Mail> {

    private final UnitPointerService unitPointerService;
    private StorylineHistoryRepository storylineRepository;
    private final Storyline<Mail> storyLine;
    private String defaultUnitName;

    public StorylineMailService(
            UnitPointerService unitPointerService,
            List<Object> unitConfigurations
    ) {
        this.storyLine = new StorylineFactory<Mail>(unitConfigurations).createStoryLine();
        this.unitPointerService = unitPointerService;
    }

    public void setStorylineHistoryRepository(StorylineHistoryRepository storylineHistoryRepository) {
        this.storylineRepository = storylineHistoryRepository;
    }

    @Override
    public Uni<Void> save(@NotNull StorylineHistory storylineHistory) {
        isNotNull(storylineHistory, storylineHistory.getUnitName(), storylineHistory.getPersonId(), storylineHistory.getMessage());
        return unitPointerService.save(new UnitPointer(storylineHistory.getPersonId(), storylineHistory.getUnitName()))
                .call(ignored -> {
                    if (checkNotNull(storylineRepository)) {
                        return storylineRepository.save(storylineHistory);
                    }
                    return Uni.createFrom().voidItem();
                }).replaceWithVoid();
    }

    @Override
    public Uni<MainUnit<Mail>> getUnitNameByPersonId(@NotNull String personId) {
        isNotNull(personId);
        return unitPointerService.getUnitNameByPersonId(personId)
                .onItem().transform(
                        unitName -> {
                            if (checkNotNull(unitName)) {
                                return storyLine.getUnit(unitName).orElse(null);
                            } else {
                                return null;
                            }
                        }
                );
    }

    @Override
    public Uni<Set<MainUnit<Mail>>> getNextUnitByPersonId(@NotNull String personId) {
        return getUnitNameByPersonId(personId)
                .flatMap(
                        unit -> {
                            if (checkNotNull(unit) && checkNotEmpty(unit.getNextUnits())) {
                                return Uni.createFrom().item(unit.getNextUnits());
                            } else {
                                return Uni.createFrom().deferred(
                                                () -> {
                                                    if (checkNotNull(storylineRepository)) {
                                                        return storylineRepository.cleanHistoryByPersonId(personId);
                                                    }
                                                    return Uni.createFrom().voidItem();
                                                }
                                        )
                                        .map(ignored -> storyLine.getStartingUnits());
                            }
                        }
                ).map(
                        nextUnits -> {
                            if (checkEmpty(nextUnits)) {
                                return storyLine.getGlobalUnits();
                            } else {
                                nextUnits.addAll(storyLine.getGlobalUnits());
                                return nextUnits;
                            }
                        }
                );
    }

    @Override
    public Uni<StorylineHistory> replaceUserToBack(String personId, int countUnitsToBack) {
        if (checkNotNull(storylineRepository)) {
            return storylineRepository.findByCountLast(personId, countUnitsToBack);
        }
        return Uni.createFrom().failure(new ConfigAppException("Ошибка. Сервис StorylineHistoryRepository не сконфигурирован."));
    }

    @Override
    public Uni<StorylineHistory> replaceUserToBack(String personId, String unitName) {
        if (checkNotNull(storylineRepository)) {
            return storylineRepository.findByCountLast(personId, unitName);
        }
        return Uni.createFrom().failure(new ConfigAppException("Ошибка. Сервис StorylineHistoryRepository не сконфигурирован."));
    }

    @Override
    public Optional<MainUnit<Mail>> getDefaultUnit() {
        if (defaultUnitName == null) return Optional.empty();
        return storyLine.getUnit(defaultUnitName);
    }

    @Override
    public void setDefaultUnit(String defaultUnit) {
        defaultUnitName = defaultUnit;
    }

    @Override
    public Uni<Void> cleanHistoryByPersonId(@NotNull String personId) {
        return Uni.createFrom().deferred(
                        () -> {
                            if (checkNotNull(storylineRepository)) {
                                return storylineRepository.cleanHistoryByPersonId(personId);
                            }
                            return Uni.createFrom().voidItem();
                        }
                )
                .call(() -> unitPointerService.removeByPersonId(personId));
    }

    //TODO [22.06.2022]: Временное решение для ленивой инициализации
    @Override
    public void lazyLink(String firstName, String secondName) {
        storyLine.link(firstName, secondName);
    }

    @Override
    public Optional<MainUnit<Mail>> getUnitByName(String unitName) {
        return storyLine.getUnit(unitName);
    }

}
