package dev.struchkov.godfather.simple.core;

import dev.struchkov.godfather.main.domain.StorylineHistory;
import dev.struchkov.godfather.main.domain.content.Message;
import dev.struchkov.godfather.main.domain.unit.UnitActiveType;
import dev.struchkov.godfather.simple.context.service.Accessibility;
import dev.struchkov.godfather.simple.context.service.ErrorHandler;
import dev.struchkov.godfather.simple.context.service.Modifiable;
import dev.struchkov.godfather.simple.core.action.ActionUnit;
import dev.struchkov.godfather.simple.core.service.StorylineService;
import dev.struchkov.godfather.simple.domain.unit.MainUnit;
import dev.struchkov.godfather.simple.domain.unit.UnitRequest;
import dev.struchkov.haiti.context.exception.NotFoundException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutorService;

import static dev.struchkov.haiti.utils.Checker.checkNotEmpty;
import static dev.struchkov.haiti.utils.Checker.checkNotNull;

public class GeneralAutoResponder<M extends Message> {

    private static final Logger log = LoggerFactory.getLogger(GeneralAutoResponder.class);

    protected final StorylineService<M> storyLineService;

    protected Map<String, List<ActionUnit>> actionUnitMap = new HashMap<>();
    protected List<Modifiable<M>> modifiable;
    protected ErrorHandler errorHandler;
    protected ExecutorService executorService;

    protected GeneralAutoResponder(StorylineService<M> storyLineService) {
        this.storyLineService = storyLineService;
    }

    public void initModifiable(List<Modifiable<M>> modifiable) {
        this.modifiable = modifiable;
    }

    public void registrationActionUnit(ActionUnit actionUnit) {
        actionUnitMap.computeIfAbsent(actionUnit.getUnitType(), k -> new ArrayList<>());
        actionUnitMap.get(actionUnit.getUnitType()).add(actionUnit);
    }

    /**
     * Позволяет установить перехватчик и обработчик исключений, возникающих при обработке юнитов.
     */
    public void setErrorHandler(ErrorHandler errorHandler) {
        this.errorHandler = errorHandler;
    }

    public void setExecutorService(ExecutorService executorService) {
        this.executorService = executorService;
    }

    public void setDefaultUnit(String unitName) {
        storyLineService.setDefaultUnit(unitName);
    }

    public void processingNewMessage(M newMessage) {
        if (checkNotNull(newMessage)) {
            if (checkNotNull(executorService)) {
                CompletableFuture.runAsync(
                        () -> processing(newMessage), executorService
                ).exceptionally(ex -> {
                    log.error(ex.getMessage(), ex);
                    return null;
                });
            } else {
                processing(newMessage);
            }
        }
    }

    public void processingNewMessages(List<M> newMessages) {
        if (checkNotEmpty(newMessages)) {
            if (checkNotNull(executorService)) {
                for (M allowedMessage : newMessages) {
                    executorService.submit(() -> processing(allowedMessage));
                }
            } else {
                for (M allowedMessage : newMessages) {
                    processing(allowedMessage);
                }
            }
        }
    }

    private void processing(M message) {
        if (checkNotEmpty(modifiable)) {
            modifiable.forEach(m -> m.change(message));
        }
        final Set<MainUnit<M>> units = storyLineService.getNextUnitByPersonId(message.getFromPersonId());
        final Optional<MainUnit<M>> optAnswer = Responder.nextUnit(message, units)
                .or(storyLineService::getDefaultUnit);
        if (optAnswer.isPresent()) {
            final MainUnit<M> answer = optAnswer.get();
            answer(UnitRequest.of(answer, message));
        }
    }

    private boolean checkPermission(Optional<Accessibility> accessibility, M message) {
        return accessibility.isEmpty() || accessibility.get().check(message);
    }

    public void answer(UnitRequest<MainUnit, M> unitRequest) {
        if (checkNotNull(errorHandler)) {
            try {
                unitRequest = getAction(unitRequest);
                activeUnitAfter(unitRequest);
            } catch (Exception e) {
                errorHandler.handle(unitRequest.getMessage(), e);
            }
        } else {
            unitRequest = getAction(unitRequest);
            activeUnitAfter(unitRequest);
        }
    }

    private UnitRequest<MainUnit, M> activeUnitAfter(UnitRequest<MainUnit, M> unitRequest) {
        final Set<MainUnit<M>> nextUnits = unitRequest.getUnit().getNextUnits();
        if (checkNotNull(nextUnits)) {
            Optional<MainUnit<M>> first = nextUnits.stream()
                    .filter(unit -> UnitActiveType.AFTER.equals(unit.getActiveType()))
                    .findFirst();
            if (first.isPresent()) {
                getAction(UnitRequest.of(first.get(), unitRequest.getMessage()));
                return activeUnitAfter(UnitRequest.of(first.get(), unitRequest.getMessage()));
            }
        }
        return unitRequest;
    }

    private UnitRequest<MainUnit, M> getAction(UnitRequest<MainUnit, M> unitRequest) {
        final MainUnit<M> unit = unitRequest.getUnit();
        final M message = unitRequest.getMessage();
        final String typeUnit = unit.getType();
        if (actionUnitMap.containsKey(typeUnit)) {
            List<ActionUnit> actionUnits = actionUnitMap.get(typeUnit);
            final Class<? extends Message> messageClass = message.getClass();
            final ActionUnit<MainUnit, M> actionUnit = getActionUnit(actionUnits, messageClass);
            if (checkNotNull(actionUnit)) {
                UnitRequest<MainUnit, M> newUnitRequest = actionUnit.action(unitRequest);
                final Optional<MainUnit<M>> optDefaultUnit = storyLineService.getDefaultUnit();
                if (!unit.isNotSaveHistory() && (optDefaultUnit.isEmpty() || !optDefaultUnit.get().equals(unit))) {
                    storyLineService.save(StorylineHistory.builder()
                            .personId(message.getFromPersonId())
                            .unitName(unit.getName())
                            .message(message)
                            .build());
                }
                final MainUnit<M> newUnit = newUnitRequest.getUnit();
                return !unit.equals(newUnit) ? getAction(newUnitRequest) : unitRequest;
            }
        }
        throw new NotFoundException("ActionUnit для типа {0} не зарегистрирован", unit.getType());
    }

    private ActionUnit<MainUnit, M> getActionUnit(List<ActionUnit> actionUnits, Class<? extends Message> messageClass) {
        for (ActionUnit actionUnit : actionUnits) {
            if (actionUnit.getMessageType().isAssignableFrom(messageClass)) {
                return actionUnit;
            }
        }
        return null;
    }

}
