package dev.struchkov.godfather.simple.core;

import dev.struchkov.godfather.main.domain.content.Message;
import dev.struchkov.godfather.main.domain.unit.KeyWord;
import dev.struchkov.godfather.simple.core.util.UnitPriorityComparator;
import dev.struchkov.godfather.simple.domain.unit.MainUnit;
import lombok.extern.slf4j.Slf4j;

import java.util.Collection;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.regex.Pattern;

import static dev.struchkov.godfather.main.core.utils.Parser.splitWords;
import static dev.struchkov.haiti.utils.Checker.checkEmpty;
import static dev.struchkov.haiti.utils.Checker.checkNotEmpty;
import static dev.struchkov.haiti.utils.Checker.checkNotNull;
import static dev.struchkov.haiti.utils.Checker.checkNull;
import static dev.struchkov.haiti.utils.Exceptions.utilityClass;
import static dev.struchkov.haiti.utils.Inspector.isNotNull;

/**
 * Реализуют основную логику автоответчика.
 *
 * @author upagge [07/07/2019]
 */
@Slf4j
public final class Responder {

    /**
     * Компаратор для сортировки Unit-ов
     */
    private static final UnitPriorityComparator UNIT_PRIORITY_COMPARATOR = new UnitPriorityComparator();

    private Responder() {
        utilityClass();
    }

    /**
     * Выбирает, какой Unit будет отдан для обработки
     *
     * @param nextUnits Множество следующих Unit-ов
     * @param message   Запрос пользователя - текстовое сообщение
     * @return Юнит, который нуждается в обработке в соответствии с запросом пользователя
     */
    public static <M extends Message> Optional<MainUnit<M>> nextUnit(M message, Collection<MainUnit<M>> nextUnits) {
        isNotNull(nextUnits);
        final Set<MainUnit<M>> searchUnit = new HashSet<>();

        if (checkNotEmpty(nextUnits)) {
            for (MainUnit<M> unit : nextUnits) {
                final String text = message.getText();
                if (checkNotNull(text)) {
                    final Set<String> unitPhrases = unit.getPhrases();
                    if (checkNotEmpty(unitPhrases) && unitPhrases.contains(text)) {
                        searchUnit.add(unit);
                    }

                    final Set<Pattern> patterns = unit.getTriggerPatterns();
                    if (checkNotEmpty(patterns)) {
                        for (Pattern pattern : patterns) {
                            if (patternReg(pattern, text)) {
                                searchUnit.add(unit);
                                break;
                            }
                        }
                    }

                    if (percentageMatch(unit, text) >= unit.getMatchThreshold()) {
                        searchUnit.add(unit);
                    }
                }

                final Predicate<M> triggerCheck = unit.getTriggerCheck();
                if (checkNotNull(triggerCheck) && triggerCheck.test(message)) {
                    searchUnit.add(unit);
                }
            }
        }

        if (searchUnit.isEmpty()) {
            for (MainUnit<M> nextUnit : nextUnits) {
                if (isNotTrigger(nextUnit)) {
                    searchUnit.add(nextUnit);
                }
            }
        }

        return searchUnit.stream().max(UNIT_PRIORITY_COMPARATOR);
    }

    private static <M extends Message> boolean isNotTrigger(MainUnit<M> nextUnit) {
        return isNotPattern(nextUnit) && isNotKeyWords(nextUnit) && isNotPhrase(nextUnit) && isNotCheck(nextUnit);
    }

    private static <M extends Message> boolean isNotCheck(MainUnit<M> unit) {
        return checkNull(unit.getTriggerCheck());
    }

    private static <M extends Message> boolean isNotPhrase(MainUnit<M> unit) {
        return checkEmpty(unit.getPhrases());
    }

    private static <M extends Message> boolean isNotPattern(MainUnit<M> unit) {
        return checkEmpty(unit.getTriggerPatterns());
    }

    private static <M extends Message> boolean isNotKeyWords(MainUnit<M> unit) {
        return checkEmpty(unit.getTriggerWords());
    }

    private static boolean patternReg(Pattern pattern, String message) {
        isNotNull(pattern);
        return message.matches(pattern.pattern());
    }

    private static <M extends Message> double percentageMatch(MainUnit<M> unit, String message) {
        final Set<KeyWord> unitKeyWords = unit.getTriggerWords();
        if (checkNotEmpty(unitKeyWords)) {
            final Set<String> messageWords = splitWords(message);
            final Set<KeyWord> intersection = getIntersection(unitKeyWords, messageWords);
            final double intersectionWeight = getIntersectionWeight(intersection);
            log.trace("Ключевые слова юнита: {} ({})", unitKeyWords, unitKeyWords.size());
            log.trace("Слова, которые прислал пользователь: {}", messageWords);
            log.trace("Пересечение: {} ({})", intersection, intersectionWeight);
            log.trace("Процент: {} Необходимо: {}", intersectionWeight / unitKeyWords.size() * 100.0, unit.getMatchThreshold());
            return (double) intersection.size() / (double) unitKeyWords.size() * 100.0;
        } else {
            return 0.0;
        }
    }

    private static double getIntersectionWeight(Set<KeyWord> intersection) {
        return intersection.stream().mapToInt(KeyWord::getImportant).sum();
    }

    private static Set<KeyWord> getIntersection(Set<KeyWord> unitKeyWords, Set<String> messageWords) {
        final Set<KeyWord> intersection = new HashSet<>();
        for (KeyWord unitKeyWord : unitKeyWords) {
            if (messageWords.contains(unitKeyWord.getWord())) {
                intersection.add(unitKeyWord);
            }
        }
        return intersection;
    }

}
