package dev.struchkov.godfather.simple.core.service;

import dev.struchkov.godfather.exception.ConfigAppException;
import dev.struchkov.godfather.main.domain.StorylineHistory;
import dev.struchkov.godfather.main.domain.UnitPointer;
import dev.struchkov.godfather.main.domain.content.Message;
import dev.struchkov.godfather.simple.context.repository.StorylineHistoryRepository;
import dev.struchkov.godfather.simple.context.service.UnitPointerService;
import dev.struchkov.godfather.simple.core.Storyline;
import dev.struchkov.godfather.simple.core.StorylineFactory;
import dev.struchkov.godfather.simple.domain.unit.MainUnit;
import dev.struchkov.haiti.utils.Inspector;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.Optional;
import java.util.Set;

import static dev.struchkov.haiti.utils.Checker.checkNotNull;
import static dev.struchkov.haiti.utils.Inspector.isNotNull;

/**
 * Отвечает за работу со сценарием в личных сообщениях с пользователем.
 */
public class StorylineMailService<T extends Message> implements StorylineService<T> {

    private final UnitPointerService unitPointerService;
    private StorylineHistoryRepository storylineRepository;
    private final Storyline<T> storyLine;
    private String defaultUnitName;

    public StorylineMailService(
            UnitPointerService unitPointerService,
            List<Object> unitConfigurations
    ) {
        this.storyLine = new StorylineFactory<T>(unitConfigurations).createStoryLine();
        this.unitPointerService = unitPointerService;
    }

    public void setStorylineRepository(StorylineHistoryRepository storylineRepository) {
        this.storylineRepository = storylineRepository;
    }

    @Override
    public void save(@NotNull StorylineHistory storylineHistory) {
        isNotNull(storylineHistory, storylineHistory.getUnitName(), storylineHistory.getPersonId(), storylineHistory.getMessage());
        unitPointerService.save(new UnitPointer(storylineHistory.getPersonId(), storylineHistory.getUnitName()));
        if (checkNotNull(storylineRepository)) {
            storylineRepository.save(storylineHistory);
        }
    }

    @Override
    public Optional<MainUnit<T>> getUnitNameByPersonId(@NotNull String personId) {
        isNotNull(personId);
        return unitPointerService.getUnitNameByPersonId(personId)
                .flatMap(storyLine::getUnit);
    }

    @Override
    public Set<MainUnit<T>> getNextUnitByPersonId(@NotNull String personId) {
        final Optional<Set<MainUnit<T>>> optMainUnits = getUnitNameByPersonId(personId)
                .map(MainUnit::getNextUnits)
                .filter(mainUnits -> !mainUnits.isEmpty());
        if (checkNotNull(storylineRepository) && optMainUnits.isEmpty()) {
            storylineRepository.cleanHistoryByPersonId(personId);
        }
        final Set<MainUnit<T>> nextUnits = optMainUnits.orElse(storyLine.getStartingUnits());
        nextUnits.addAll(storyLine.getGlobalUnits());
        return nextUnits;
    }

    @Override
    public Optional<StorylineHistory> replaceUserToBack(String personId, int countUnitsToBack) {
        Inspector.isNotNull(() -> new ConfigAppException("Ошибка. Сервис StorylineHistoryRepository не сконфигурирован."), storylineRepository);
        return storylineRepository.findByCountLast(personId, countUnitsToBack);
    }

    @Override
    public Optional<StorylineHistory> replaceUserToBack(String personId, String unitName) {
        Inspector.isNotNull(() -> new ConfigAppException("Ошибка. Сервис StorylineHistoryRepository не сконфигурирован."), storylineRepository);
        return storylineRepository.findByCountLast(personId, unitName);
    }

    @Override
    public Optional<MainUnit<T>> getDefaultUnit() {
        if (defaultUnitName == null) return Optional.empty();
        return storyLine.getUnit(defaultUnitName);
    }

    @Override
    public void setDefaultUnit(String defaultUnit) {
        defaultUnitName = defaultUnit;
    }

    @Override
    public void cleanHistoryByPersonId(@NotNull String personId) {
        if (checkNotNull(storylineRepository)) {
            storylineRepository.cleanHistoryByPersonId(personId);
        }
        unitPointerService.removeByPersonId(personId);
    }

    //TODO [22.06.2022]: Временное решение для ленивой инициализации
    @Override
    public void lazyLink(String firstName, String secondName) {
        storyLine.link(firstName, secondName);
    }

    @Override
    public Optional<MainUnit<T>> getUnitByName(String unitName) {
        return storyLine.getUnit(unitName);
    }

}
