package dev.struchkov.godfather.quarkus.domain;

import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import dev.struchkov.godfather.main.domain.jackson.TelegramPayloadDeserializer;
import dev.struchkov.godfather.main.domain.jackson.TelegramPayloadSerializer;
import dev.struchkov.godfather.main.domain.keyboard.KeyBoard;
import dev.struchkov.godfather.quarkus.domain.content.send.SendAttachment;
import dev.struchkov.haiti.utils.container.ContextKey;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;
import lombok.ToString;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import static dev.struchkov.haiti.utils.Checker.checkNotNull;
import static dev.struchkov.haiti.utils.Checker.checkNull;

/**
 * Контейнер, которые содержит данные, которые будут отправлены пользователю как ответ на его запрос.
 *
 * @author upagge [08/07/2019]
 */
@Getter
@Setter
@ToString
@NoArgsConstructor
public class BoxAnswer {

    /**
     * <p>Получатель сообщения</p>
     */
    private String recipientPersonId;

    /**
     * Идентификатор сообщения, которое нужно заменить.
     */
    private String replaceMessageId;

    /**
     * Идентификатор сообщения, на которое отвечаем.
     */
    private String replyToMessageId;

    /**
     * Клавиатура - меню.
     */
    private KeyBoard keyBoard;

    /**
     * Флаг означающий, что надо перезаписать наше последнее отправленное сообщение, вместо отправки нового.
     */
    private boolean replace;

    /**
     * Обычное текстовое сообщение.
     */
    private String message;

    /**
     * Полезная нагрузка для реализаций.
     */
    @JsonSerialize(using = TelegramPayloadSerializer.class)
    @JsonDeserialize(using = TelegramPayloadDeserializer.class)
    protected Map<ContextKey, Object> payload = new HashMap<>();

    private SendAttachment attachment;

    private BoxAnswer(Builder builder) {
        message = builder.message;
        keyBoard = builder.keyBoard;
        replace = builder.replace;
        replaceMessageId = builder.replaceMessageId;
        recipientPersonId = builder.recipientPersonId;
        payload = builder.payload;
        attachment = builder.attachment;
        replyToMessageId = builder.replyToMessageId;
    }

    public static BoxAnswer boxAnswer(boolean replace, String message) {
        return BoxAnswer.builder().replace(replace).message(message).build();
    }

    public static BoxAnswer boxAnswer(boolean replace, String messageText, KeyBoard keyBoard) {
        return BoxAnswer.builder().replace(replace).message(messageText).keyBoard(keyBoard).build();
    }

    public static BoxAnswer boxAnswer(String message) {
        return boxAnswer(false, message);
    }

    public static BoxAnswer boxAnswer(String message, KeyBoard keyBoard) {
        return boxAnswer(false, message, keyBoard);
    }

    public static BoxAnswer replaceBoxAnswer(String message) {
        return boxAnswer(true, message);
    }

    public static BoxAnswer replaceBoxAnswer(String message, KeyBoard keyBoard) {
        return boxAnswer(true, message, keyBoard);
    }

    public static BoxAnswer replaceBoxAnswer(String messageId, String message) {
        return BoxAnswer.builder().replace(true).replaceMessageId(messageId).message(message).build();
    }

    public static BoxAnswer replaceBoxAnswer(String messageId, String message, KeyBoard keyBoard) {
        return BoxAnswer.builder().replace(true).replaceMessageId(messageId).message(message).keyBoard(keyBoard).build();
    }

    public static Builder builder() {
        return new Builder();
    }

    public static Builder replaceBuilder() {
        return new Builder().replace(true);
    }

    public void setRecipientPersonId(String recipientPersonId) {
        this.recipientPersonId = recipientPersonId;
    }

    public void setRecipientIfNull(String recipientPersonId) {
        if (checkNull(this.recipientPersonId)) {
            this.recipientPersonId = recipientPersonId;
        }
    }

    public Map<ContextKey, Object> getPayload() {
        return payload;
    }

    public <T> void setPayload(ContextKey<T> key, T value) {
        if (checkNotNull(value)) {
            payload.put(key, value);
        }
    }

    public <T> Optional<T> getPayLoad(ContextKey<T> contextKey) {
        if (checkNotNull(payload)) {
            return Optional.ofNullable(payload.get(contextKey))
                    .map(value -> (T) value);
        }
        return Optional.empty();
    }

    public static final class Builder {

        private String message;
        private String replyToMessageId;
        private KeyBoard keyBoard;
        private boolean replace;
        private String replaceMessageId;
        private String recipientPersonId;
        private Map<ContextKey, Object> payload = new HashMap<>();
        private SendAttachment attachment;

        private Builder() {
        }

        public Builder message(String val) {
            message = val;
            return this;
        }

        public Builder replyToMessageId(String val) {
            replyToMessageId = val;
            return this;
        }

        public Builder keyBoard(KeyBoard val) {
            keyBoard = val;
            return this;
        }

        public Builder replace(boolean val) {
            replace = val;
            return this;
        }

        public Builder recipientPersonId(String val) {
            recipientPersonId = val;
            return this;
        }

        public Builder replaceMessageId(String val) {
            replaceMessageId = val;
            return this;
        }

        public <T> Builder payload(ContextKey<T> key, T value) {
            if (checkNotNull(value)) {
                payload.put(key, value);
            }
            return this;
        }

        public Builder payload(ContextKey<Boolean> key) {
            if (checkNotNull(key)) {
                payload.put(key, true);
            }
            return this;
        }

        public Builder attachment(SendAttachment attachment) {
            if (checkNotNull(attachment)) {
                this.attachment = attachment;
            }
            return this;
        }

        public BoxAnswer build() {
            return new BoxAnswer(this);
        }

    }

}
