package dev.struchkov.godfather.quarkus.domain.unit;

import dev.struchkov.godfather.main.domain.content.Message;
import dev.struchkov.godfather.main.domain.unit.KeyWord;
import dev.struchkov.godfather.main.domain.unit.UnitActiveType;
import dev.struchkov.godfather.quarkus.domain.BoxAnswer;
import dev.struchkov.godfather.quarkus.domain.unit.func.CallBackConsumer;
import dev.struchkov.godfather.quarkus.domain.unit.func.ProcessingData;
import dev.struchkov.godfather.quarkus.domain.unit.func.UniConsumer;
import dev.struchkov.godfather.quarkus.domain.unit.func.UniPredicate;
import io.smallrye.mutiny.Uni;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import static dev.struchkov.godfather.quarkus.domain.BoxAnswer.boxAnswer;

/**
 * Используется для отправки ответа пользователю.
 *
 * @author upagge [08/07/2019]
 */
public class AnswerText<M extends Message> extends MainUnit<M> {

    public static final String TYPE = "ANSWER_TEXT";

    /**
     * Объект, который необходимо отправить пользователю.
     */
    private final ProcessingData<M> answer;

    /**
     * Объект нестандартной отправки ответа.
     */
    private final CallBackConsumer callBack;

    private AnswerText(Builder<M> builder) {
        super(
                builder.name,
                builder.triggerWords,
                builder.triggerPhrases,
                builder.triggerPatterns,
                builder.triggerCheck,
                builder.matchThreshold,
                builder.priority,
                builder.nextUnits,
                builder.description,
                TYPE,
                builder.activeType,
                builder.notSaveHistory
        );
        answer = builder.boxAnswer;
        callBack = builder.callBack;
    }

    public static <M extends Message> AnswerText<M> of(String message) {
        return AnswerText.<M>builder().answer(() -> boxAnswer(message)).build();
    }

    public static <M extends Message> Builder<M> builder() {
        return new Builder<>();
    }

    public ProcessingData<M> getAnswer() {
        return answer;
    }

    public CallBackConsumer getCallBack() {
        return callBack;
    }

    public static final class Builder<M extends Message> {

        private String name = UUID.randomUUID().toString();
        private String description;
        private Set<MainUnit<M>> nextUnits;

        private Set<KeyWord> triggerWords;
        private Set<String> triggerPhrases;
        private UniPredicate<M> triggerCheck;
        private Set<Pattern> triggerPatterns;

        private Integer matchThreshold;
        private Integer priority;

        private UnitActiveType activeType;
        private boolean notSaveHistory;

        private ProcessingData<M> boxAnswer;
        private CallBackConsumer callBack;

        private Builder() {
        }

        public Builder<M> name(String name) {
            this.name = name;
            return this;
        }

        public Builder<M> description(String description) {
            this.description = description;
            return this;
        }

        public Builder<M> answerConsumer(Consumer<M> consumer) {
            this.boxAnswer = message -> Uni.createFrom().item(() -> {
                consumer.accept(message);
                return null;
            });
            return this;
        }

        public Builder<M> answerConsumerUni(UniConsumer<M> uniConsumer) {
            this.boxAnswer = message -> uniConsumer.accept(message).replaceWith(() -> null);
            return this;
        }

        public Builder<M> answerUni(Uni<?> answer) {
            this.boxAnswer = message -> answer.replaceWith(() -> null);
            return this;
        }

        public Builder<M> answerUni(Function<M, Uni<BoxAnswer>> answer) {
            this.boxAnswer = answer::apply;
            return this;
        }

        public Builder<M> answer(Function<M, BoxAnswer> answer) {
            this.boxAnswer = message -> Uni.createFrom().item(() -> answer.apply(message));
            return this;
        }

        public Builder<M> answer(Supplier<BoxAnswer> answer) {
            this.boxAnswer = message -> Uni.createFrom().item(answer);
            return this;
        }

        public Builder<M> answerUni(Supplier<Uni<BoxAnswer>> answer) {
            this.boxAnswer = message -> answer.get();
            return this;
        }

        public Builder<M> callBack(CallBackConsumer callBack) {
            this.callBack = callBack;
            return this;
        }

        public Builder<M> triggerWords(Set<KeyWord> val) {
            if (triggerWords == null) {
                triggerWords = new HashSet<>();
            }
            triggerWords.addAll(val);
            return this;
        }

        public Builder<M> triggerWord(KeyWord val) {
            if (triggerWords == null) {
                triggerWords = new HashSet<>();
            }
            triggerWords.add(val);
            return this;
        }

        public Builder<M> triggerStringWords(Set<String> val) {
            if (triggerWords == null) {
                triggerWords = new HashSet<>();
            }
            triggerWords.addAll(val.stream().map(KeyWord::of).collect(Collectors.toSet()));
            return this;
        }

        public Builder<M> triggerWord(String val) {
            if (triggerWords == null) {
                triggerWords = new HashSet<>();
            }
            triggerWords.add(KeyWord.of(val));
            return this;
        }

        public Builder<M> triggerPhrase(String... val) {
            if (triggerPhrases == null) {
                triggerPhrases = new HashSet<>();
            }
            if (val.length == 1) {
                triggerPhrases.add(val[0]);
            } else {
                triggerPhrases.addAll(Set.of(val));
            }
            triggerPhrases.addAll(List.of(val));
            return this;
        }

        public Builder<M> triggerPattern(Pattern... val) {
            if (triggerPatterns == null) {
                triggerPatterns = new HashSet<>();
            }
            if (val.length == 1) {
                triggerPatterns.add(val[0]);
            } else {
                triggerPatterns.addAll(Set.of(val));
            }
            triggerPatterns.addAll(Set.of(val));
            return this;
        }

        public Builder<M> triggerCheckUni(UniPredicate<M> trigger) {
            triggerCheck = trigger;
            return this;
        }

        public Builder<M> triggerCheck(Predicate<M> trigger) {
            triggerCheck = UniPredicate.predicate(trigger);
            return this;
        }

        public Builder<M> matchThreshold(Integer val) {
            matchThreshold = val;
            return this;
        }

        public Builder<M> priority(Integer val) {
            priority = val;
            return this;
        }

        public Builder<M> next(MainUnit<M> val) {
            if (nextUnits == null) {
                nextUnits = new HashSet<>();
            }
            nextUnits.add(val);
            return this;
        }

        public Builder<M> notSaveHistory() {
            notSaveHistory = true;
            return this;
        }

        public Builder<M> activeType(UnitActiveType val) {
            activeType = val;
            return this;
        }

        public AnswerText<M> build() {
//            isNotNull(boxAnswer, UnitConfigException.unitConfigException("BoxAnswer обязательный параметр юнита"));
            return new AnswerText<>(this);
        }

    }

}
